/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: ToneGen.cpp,v $
   Revision 1.1  2007/12/01 06:40:59  wb6ymh
   Inital import - general purpose tone generator class.


*/
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "ToneGen.h"

#ifndef  M_PI
#define  M_PI     3.14159265358979323846  /* pi */
#endif

CToneGen::CToneGen()
{
   bToneGenActive = TRUE;
   pToneGenSpecStart = NULL;
   pToneGen = NULL;
   Tone1Level = 8000;
   Tone2Level = 8000;
   Tone1EndFreq = 0;
   ToneDuration = 0;
}

// Generate a tones based on compiled tone specification
// return the number of samples in ToneBuf
int CToneGen::GenSamples(int16 *ToneBuf,int i,int Samples)
{
   if(bToneGenActive && ToneGenSamples == 0) {
      ToneGenSetup();
   }

   D2PRINTF(("%s: bToneGenActive: %d, Tone1Freq: %d, Tone2Freq: %d, ToneGenSamples: %d\n",
             __FUNCTION__,bToneGenActive,Tone1Freq,Tone2Freq,ToneGenSamples));

   if(Tone1Freq != 0 && Tone2Freq != 0) {
   // Both tones running
      for(; i < Samples; i++) {
         if(ToneGenSamples == 0) {
            break;
         }
         ToneGenSamples--;
         ToneBuf[i] = (int16) (Tone1Level * sin(Tone1Angle) + 
                               Tone2Level * sin(Tone2Angle));
         Tone1Angle += DeltaAngle1;
         DeltaAngle1 += DeltaAngle1Delta;
         Tone2Angle += DeltaAngle2;
      }
   }
   else if(Tone1Freq != 0) {
   // Just tone 1 running
      for(; i < Samples; i++) {
         if(ToneGenSamples == 0) {
            break;
         }
         ToneGenSamples--;
         ToneBuf[i] = (int16) (Tone1Level * sin(Tone1Angle));
         Tone1Angle += DeltaAngle1;
         DeltaAngle1 += DeltaAngle1Delta;
      }
   }
   else if(Tone2Freq != 0) {
   // Just tone 2 running
      for(; i < Samples; i++) {
         if(ToneGenSamples == 0) {
            break;
         }
         ToneGenSamples--;
         ToneBuf[i] = (int16) (Tone2Level * sin(Tone2Angle));
         Tone2Angle += DeltaAngle2;
      }
   }
   else {
   // No tones
      int Samples2Copy;
      int j;

      if(bToneGenActive) {
         Samples2Copy = ToneGenSamples;
         if(Samples2Copy > Samples - i) {
            Samples2Copy = Samples - i;
         }
         ToneGenSamples -= Samples2Copy;
      }
      else {
         Samples2Copy = Samples - i;
      }

      if(LastToneSample != 0) {
      // Ramp the last tone sample out to prevent thumps
         for(j = 0; j < Samples2Copy; j++) {
            if(LastToneSample == 0) {
               Tone1Angle = Tone2Angle = 0.0;
               break;
            }
            LastToneSample /= 2;
            ToneBuf[i++] = LastToneSample;
         }
         Samples2Copy -= j;
      }

      if(Samples2Copy > 0) {
         memset(&ToneBuf[i],0,Samples2Copy * sizeof(int16));
         i += Samples2Copy;
      }
   }
   return i;
}

int CToneGen::ToneGen(int16 *ToneBuf,int Samples)
{
   int i = 0;
   int Ret = 0;

   while(i < Samples) {
      i = GenSamples(ToneBuf,i,Samples);
   }
// Save last tone sample for rampout
   LastToneSample = ToneBuf[Samples-1];
   if(!bToneGenActive && LastToneSample == 0) {
   // We're done
      Ret = TRUE;
   }
   D2PRINTF(("%s: returning %d\n",__FUNCTION__,Ret));
   return Ret;
}

int CToneGen::ToneGenSetup(char *pToneGenSpec,char **pErrPos)
{
   char *cp;
   char Tag[4];
   int x;
   int ParseErrorLine = 0;
   int i;
   char *pParseStart;

// Set defaults
   if(pToneGenSpec != NULL) {
      ToneRepCount = -1;
      pToneGenLastRep = pToneGen = pToneGenSpecStart = strdup(pToneGenSpec);
   }

   cp = pToneGen;

   bToneGenActive = FALSE;
   Tone1Freq = 0;
   Tone2Freq = 0;
   Tone1EndFreq = 0;
   ToneDuration = 0;

   if(*cp == 0) {
   // End of the spec, we're done
      free(pToneGenSpecStart);
      pToneGenSpecStart = NULL;
      pToneGen = NULL;
      bToneGenActive = FALSE;
   }
   else {
      while(*cp) {
         while(*cp == ' ') {
            cp++;
         }

         if(*cp == 0) {
         // that's it folks!
            break;
         }

         pParseStart = cp;

      // copy the command tag
         for(i = 0; i < 3; i++) {
            if(!isprint(*cp) || *cp == ' ') {
               break;
            }
            Tag[i] = *cp++;
         }
         Tag[i] = 0;

         while(*cp == ' ') {
            cp++;
         }

         if(sscanf(cp,"%d",&x) != 1) {
         // Parse error
            ParseErrorLine = __LINE__;
            break;
         }

         while(*cp != 0 && *cp != ' ') {
            cp++;
         }

         pToneGen = cp;

         if(STRCMP(Tag,"TF1") == 0) {
            Tone1Freq = x;
         }
         else if(STRCMP(Tag,"EF1") == 0) {
            Tone1EndFreq = x;
         }
         else if(STRCMP(Tag,"TL") == 0) {
            Tone1Level = Tone2Level = x;
         }
         else if(STRCMP(Tag,"TF2") == 0) {
            Tone2Freq = x;
         }
         else if(STRCMP(Tag,"TL2") == 0) {
            Tone2Level = x;
         }
         else if(STRCMP(Tag,"DUR") == 0) {
            ToneDuration = x;
            break;
         }
         else if(STRCMP(Tag,"RPT") == 0) {
            if(ToneRepCount == -1) {
            // Rep count hasn't been set yet
               ToneRepCount = x;
            }

            if(--ToneRepCount > 0) {
            // Repeat count not satisfied  yet, do it all again
               pToneGen = pToneGenLastRep;
               ToneGenSetup();
            }
            else {
            // Finished with this group, setup for the next group
               pToneGenLastRep = pToneGen;
               ToneRepCount = -1;
               ToneGenSetup();
            }
            break;
         }
         else if(STRCMP(Tag,"ID") == 0) {
            ID = x;
         }
         else {
            ParseErrorLine = __LINE__;
            break;
         }
      }

      if(ParseErrorLine != 0) {
         LOG_ERROR(("%s#%d: unable to parse tonegen spec \"%s\"\n",__FUNCTION__,
                    ParseErrorLine,pParseStart));
         if(pErrPos != NULL) {
            *pErrPos = pToneGenSpec + (pParseStart - pToneGenSpecStart);
         }
         bToneGenActive = FALSE;
         Tone1Freq = 0;
         Tone2Freq = 0;
         Tone1EndFreq = 0;
         ToneDuration = 0;
      }
      else if(ToneDuration != 0) {
      // There's something to do 
         bToneGenActive = TRUE;

      // calculate number of samples for the specified duration
         ToneGenSamples = (ToneDuration * 8000) / 1000;

         DeltaAngle1 = 2 * M_PI * Tone1Freq / 8000.0;
         DeltaAngle2 = 2 * M_PI * Tone2Freq / 8000.0;
         if(Tone1EndFreq != 0) {
         // Setup tone sweep
            DeltaAngle1Delta = ((2 * M_PI * Tone1EndFreq / 8000.0) 
                               - DeltaAngle1) / ToneGenSamples;
         }
         else {
            DeltaAngle1Delta = 0.0;
         }
      }

      if(!bToneGenActive) {
         if(pToneGenSpecStart != NULL) {
            free(pToneGenSpecStart);
            pToneGenSpecStart = NULL;
         }
      }
   }
// Go nuke yourself !
   if(ParseErrorLine != 0) {
      delete this;   
   }
   return ParseErrorLine;
}

